from typing import List, Tuple
import json
from hmac import HMAC
from hashlib import sha256

from trezorlib.misc import encrypt_keyvalue
from trezorlib.tools import parse_path
from trezorlib.client import TrezorClient
from trezorlib.exceptions import TrezorException, TrezorFailure

from ..crypto import PATH, FILENAME_MESS, decrypt
from .entry import Entry
from .sources import Source
from .tag import Tag


class Store:
    def __init__(self, client: TrezorClient, source: Source):
        self.client = client
        self.entries: List[Entry] = []
        self.tags: List[Tag] = []
        self._master_key = self._load_master_key(client)
        self.source = source

    async def load(self) -> None:
        """Loads the store content from the source

        Raises:
            StoreLoadError
            StoreDecryptError
            StoreDecodeError
        """
        encrypted_store = await self._load_encrypted_store()
        encoded_store = self._decrypt_store(encrypted_store)
        (tags, entries) = self._decode_store(encoded_store)
        self.tags = tags
        self.entries = entries

    async def __aenter__(self):
        await self.load()
        return self

    async def __aexit__(self, *args):
        pass

    @staticmethod
    def _load_master_key(client: TrezorClient) -> str:
        address_n = parse_path(PATH)
        key = "Activate TREZOR Password Manager?"
        value = bytes.fromhex(
            "2d650551248d792eabf628f451200d7f51cb63e46aadcbb1038aacb05e8c8aee2d650551248d792eabf628f451200d7f51cb63e46aadcbb1038aacb05e8c8aee")
        try:
            return encrypt_keyvalue(client, address_n, key, value).hex()
        except TrezorException:
            raise
        except:
            raise TrezorFailure()

    @property
    def name(self) -> str:
        file_key = self._master_key[:len(self._master_key) // 2]
        return HMAC(file_key.encode(), FILENAME_MESS.encode(), sha256).hexdigest() + '.pswd'

    async def _load_encrypted_store(self) -> bytes:
        try:
            return await self.source.load_store(self.name)
        except Exception as e:
            raise StoreLoadError() from e

    def _decrypt_store(self, encrypted_store: bytes) -> bytes:
        """Returns JSON representation of the password store"""
        enc_key = self._master_key[len(self._master_key) // 2:]
        try:
            return decrypt(enc_key, encrypted_store)
        except Exception as e:
            raise StoreDecryptError() from e

    @staticmethod
    def _decode_store(json_store: bytes) -> Tuple[List[Tag], List[Entry]]:
        """Returns object representation of the password store"""
        try:
            dict = json.loads(json_store)
            raw_tags = {int(key): Tag.load(dict['tags'][key]) for key in dict['tags']}
            tags = [raw_tags[key] for key in raw_tags]
            entries = [Entry.load(dict['entries'][key], raw_tags) for key in dict['entries']]
            return [tags, entries]
        except Exception as e:
            raise StoreDecodeError() from e


class StoreLoadError(Exception):
    pass


class StoreDecryptError(Exception):
    pass


class StoreDecodeError(Exception):
    pass



import asyncio

from trezorlib.client import TrezorClient
from trezorlib.transport import TransportException

from ..client import get_safe_client
from ..utils import animate_dots, prompt_print


healthcheck_tasks = set()


async def get_client() -> TrezorClient:
    """Waits for the Trezor client to connect

    Returns:
        The connected Trezor client

    Raises:
        KeyboardInterrupt
        Exception: Client initialization failure
    """
    prompt_print("Looking for a Trezor device ", end="", flush=True)
    animation = animate_dots(5)
    while True:
        try:
            client = get_safe_client()
            task = asyncio.get_event_loop().create_task(client_healthcheck(client))
            healthcheck_tasks.add(task)
            print()
            return client
        except TransportException:
            next(animation)  # Keep waiting
        except Exception:
            print()
            print("Unable to access a Trezor device")
            raise
        await asyncio.sleep(0.7)


async def client_healthcheck(client: TrezorClient):
    while True:
        try:
            await asyncio.get_event_loop().run_in_executor(None, client.ping, "healthcheck")
        except Exception as e:
            for task in asyncio.tasks.all_tasks():
                if not task.cancelled():
                    task.cancel("Trezor device has been disconnected")
            return
        await asyncio.sleep(1)


import threading

from trezorlib.client import TrezorClient
from trezorlib.transport import Transport, get_transport

from .ui import ManagerUI


def get_safe_client():
    """Creates default thread-safe client instance

    Returns:
        Instance of the thread-safe Trezor client

    Raises:
        TransportException: when client cannot be found
        Exception: when client initialization fails
    """
    transport = get_transport()
    ui = ManagerUI()
    return SafeTrezorClient(transport, ui)


class SafeTrezorClient(TrezorClient):
    def __init__(self, transport: Transport, ui: ManagerUI, **kwargs):
        self._lock = threading.Lock()
        super().__init__(transport, ui, _init_device=False, **kwargs)

    def call_raw(self, msg):
        with self._lock:
            return super().call_raw(msg)

    def __enter__(self):
        self.init_device()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end_session()